(***************************************)
(*** С 饤ȥ⥸塼 ***)
(***************************************)

(* $Id: rvclient.ml 8 2005-05-10 01:47:07Z mame $ *)

(***  ********************************************************************)
open Arg
let debug_r = ref false         (* true ʤ̿ƤϤ *)
let host_r = ref "localhost"    (* ³ۥ̾ *)
let port_r = ref 10000          (* ³ݡ *)
let id_r = ref ""               (* 饤 ID *)
let console_r = ref false       (* 󥽡⡼ *)
let _ =
    parse [
        ("-h", Set_string host_r,
            "Specify a host to connect over the network");
        ("-p", Set_int port_r, "Specify a port number to connect");
        ("-i", Set_string id_r, "Specify id");
        ("-d", Set debug_r, "Debug mode");
        ("-c", Set console_r, "Console mode")]
        (fun x -> raise (Bad (x ^ ": parse error")))
        "Usage: ./mak [-h HOST] [-p PORT] [-i ID]"
let debug = !debug_r
let host = !host_r
let port = !port_r
let id = !id_r
let console = !console_r

(***  ********************************************************************)
type event =
| Game_start of string
| Game_end of bool
| Set_start of (Reversi.color * int)
| Set_end of (int * int)
| Next of state
and state = {
    board: Reversi.stone list list;
    color: Reversi.color;
    rest: int;
    history: (Reversi.color * int * int) list;
}
exception ProtocolError of string

(*** Хѿ **********************************************************)
let in_c_r = ref None   (* Ф *)
and out_c_r = ref None  (* Фؤν *)
and id_r = ref None     (* ʬ ID *)

(*** utils *******************************************************************)
(* ̤ʸɽ stone list list Ѵ *)
(* "---------------------------XO------OX---------------------------" *)
let stone_list_list_of_string str =
    let board = Array.init 8 (fun _ -> Array.make 8 None) in
    for y = 0 to 7 do
        for x = 0 to 7 do
            board.(y).(x) <-
                match str.[y * 8 + x] with
                | 'X' -> Some Reversi.Black
                | 'O' -> Some Reversi.White
                | _ -> None;
        done;
    done;
    Array.to_list (Array.map (fun a -> Array.to_list a) board)
(* Фοʸɽ Reversi.color Ѵ *)
(* "X" : Black, "O" : White *)
let color_of_string str =
    match str with
    | "X" -> Reversi.Black
    | "O" -> Reversi.White
    | _ -> failwith "malformed stone."
(* ʸɽ (Reversi.color * int * int) list Ѵ *)
(* "[XC4OC5XE6]" *)
let history_of_string str =
    let func l =
        let clr = color_of_string (String.sub str (l * 3 + 1) 1) in
        let x = (Char.code str.[l * 3 + 2]) - 65 (* 'A' *) in
        let y = (Char.code str.[l * 3 + 3]) - 49 (* '1' *) in
        (clr, x, y)
    in Array.to_list (Array.init (((String.length str) - 2) / 3) func)

open Unix

(*** ᥤץ ********************************************************)
(*  (³ ID ) *)
let init id' =
    let host =
        try
            (gethostbyname host).h_addr_list.(0)
        with _ ->
            inet_addr_of_string host
    in
    let id = String.uppercase (if id = "" then id' else id) in
    let (in_c, out_c) =
        if console then
            (Pervasives.stdin, Pervasives.stdout)
        else
            open_connection (ADDR_INET (host, port))
    in
    in_c_r := Some in_c;
    out_c_r := Some out_c;
    id_r := Some id;
    output_string out_c ("2 " ^ id ^ "\n");
    flush out_c

(* Υ٥ȼ *)
let rec get () =
    let (in_c, out_c) =
        match (!in_c_r, !out_c_r) with
        | (Some in_c, Some out_c) -> (in_c, out_c)
        | _ -> failwith "not initialized."
    in
    let line1 = String.uppercase (input_line in_c) in
    let line2 = String.uppercase (input_line in_c) in
    if debug then print_endline ("S: " ^ line1);
    if debug then print_endline ("S: " ^ line2);
    flush Pervasives.stdout;
    try
        begin
            match line1 with
            | "START_GAME" ->
                output_string out_c "\n"; flush out_c;
                Game_start line2
            | "END_GAME" ->
                output_string out_c "\n"; flush out_c;
                Game_end (!id_r = Some line2)
            | "START" ->
                output_string out_c "\n"; flush out_c;
                let func clr rst = Set_start (color_of_string clr, rst) in
                Scanf.sscanf line2 "%s %d" func
            | "END" ->
                output_string out_c "\n"; flush out_c;
                let func slf opp = Set_end (slf, opp) in
                Scanf.sscanf line2 "%d %d" func
            | "NEXT" ->
                let func brd clr hty rst =
                    Next {
                        board = stone_list_list_of_string brd;
                        color = color_of_string clr;
                        history = history_of_string hty;
                        rest = rst;
                    }
                in
                Scanf.sscanf line2 "%s %s %s %d" func
            | _ ->
                raise
                    (ProtocolError
                        ("invalid command: " ^ line1 ^ ": " ^ line2))
        end
    with Scanf.Scan_failure _ | Failure _ | End_of_file ->
        raise (ProtocolError ("malformed " ^ line1 ^ ": " ^ line2))


(* Ф֤ *)
let put x y =
    let out_c =
        match !out_c_r with
        | Some out_c -> out_c
        | None -> failwith "not initialized."
    in
    let s = "  \n" in
    s.[0] <- "ABCDEFGH".[x];
    s.[1] <- "12345678".[y];
    if debug then print_string ("C: " ^ s);
    output_string out_c s;
    flush out_c

(*

***** ץȥ복 *****

C: <ץȥС = 2> <ʬID>

S: START_GAME
S: <ID>
C:

S: START
S: <ʬФο> <>
C:

S: NEXT
S: <> <ʬФο> <> <Ĥ>
C: <>
...

S: END
S: <ʬФο> <ŨФο>
C:
...

S: END_GAME
S: <äץ쥤䡼 ID>



***** ץȥμ¹ *****

C: 2 HOGE

S: START_GAME
S: FUGA
C:

S: START
S: X 300000
C:

S: NEXT
S: ---------------------------OX------XO--------------------------- X [] 300000
C: C4

S: NEXT
S: --------------------------XXX-----OOO--------------------------- X [XC4OC5] 299996
C: E6

S: NEXT
S: ------------------O-------OXX-----OXX-------X------------------- X [XC4OC5XE6OC3] 299993
C: B2
...

S: END
S: 38 26
C:

S: START
S: O
C:

S: NEXT
S: ---------------------------OX------XXX-------------------------- O [XF5] 300000
C: D6
...

S: END_GAME
S: HOGE


*)
