(*********************************)
(*** С Хץ ***)
(*********************************)

(* $Id: rvserver.ml 8 2005-05-10 01:47:07Z mame $ *)

(***  ********************************************************************)
open Arg
let port_r = ref 10000                  (* Ԥݡ *)
let number_of_sets_r = ref 4            (* åȿ (ˤ٤) *)
let time_limit_1_r = ref ( 30 * 1000)   (* Ⱦ 2 åȤ» (ms) *)
let time_limit_2_r = ref (300 * 1000)   (* Ⱦ 2 åȤ» (ms) *)
let print_each_board_r = ref false      (* ꤴȤ̤ɽ *)
let process1_r = ref ""                 (* ץ쥤䡼 1 ¹ԥեѥ *)
let process2_r = ref ""                 (* ץ쥤䡼 2 ¹ԥեѥ *)
let _ =
    parse [
        ("-p", Set_int port_r, "Specify a port number to listen");
        ("-c", Set_int number_of_sets_r, "Specify the number of sets");
        ("-t1", Set_int time_limit_1_r, "Specify time limit 1");
        ("-t2", Set_int time_limit_2_r, "Specify time limit 2");
        ("-b", Set print_each_board_r, "Print each board");
        ("-p1", Set_string process1_r,
            "Specify the executable filename of player 1");
        ("-p2", Set_string process2_r,
            "Specify the executable filename of player 2")]
        (fun x -> raise (Bad (x ^ ": parse error")))
        "Usage:
 ./rvserver [-p PORT] [-c COUNT] [-t1 TIME_LIMIT_1] [-t2 TIME_LIMIT_2]
 ./rvserver -p1 PLAYER1EXE -p2 PLAYER2EXE"
let port = !port_r
let number_of_sets = !number_of_sets_r
let time_limit_1 = !time_limit_1_r
let time_limit_2 = !time_limit_2_r
let print_each_board = !print_each_board_r
let process1 = !process1_r
let process2 = !process2_r

(***  ********************************************************************)
open Reversi
open Unix
open Sys
type handle = {
    mutable id: string;
    idx: int;
    start_game: string -> unit;
    end_game: string -> unit;
    start_set: color -> int -> unit;
    end_set: int -> int -> unit;
    next: stone list list -> color -> (color * int * int) list -> int ->
        int * int;
    close: unit -> unit;
}

(*** utils *******************************************************************)
(* stone list list ̤ʸɽѴ *)
let string_of_board brd =
    let func z e =
        let e =
            match e with
            | None -> "-"
            | Some Black -> "X"
            | Some White -> "O"
        in z ^ e
    in List.fold_left func "" (List.flatten brd)
(* Reversi.color ФοʸɽѴ *)
let string_of_color = function Black -> "X" | White -> "O"
(* (Reversi.color * int * int) list ʸɽ "[XC4OC5XE6]" Ѵ *)
let string_of_history hty =
    let func z (clr, x, y) =
        let s = "  " in
        s.[0] <- "ABCDEFGH".[x];
        s.[1] <- "12345678".[y];
        z ^ (string_of_color clr) ^ s
    in (List.fold_left func "[" (List.rev hty)) ^ "]"
(* "A1" ʤɤ int * int Ѵ *)
let int_x_int_of_move str =
    let str = String.uppercase str in
    let x = (Char.code str.[0]) - 65 (* 'A' *) in
    let y = (Char.code str.[1]) - 49 (* '1' *) in
    if x < 0 || x > 7 || y < 0 || y > 7 then failwith "illigal move.";
    (x, y)
(* board  stone list list Ѵ *)
let stone_list_list_of_board brd =
    let board = Array.init 8 (fun _ -> Array.make 8 None) in
    for y = 0 to 7 do
        for x = 0 to 7 do
            board.(y).(x) <- get_stone brd x y
        done
    done;
    Array.to_list (Array.map (fun a -> Array.to_list a) board)
(* ̤ɽ *)
let print_board brd tx ty =
    print_string "   A B C D E F G H";
    print_newline ();
    let f = ref false in
    for y = 0 to 7 do
        Printf.printf "% 2d" (y + 1);
        for x = 0 to 7 do
            if not !f then
                print_string (if x = tx && y = ty then "[" else " ");
            begin
                match get_stone brd x y with
                | None -> print_string "."
                | Some Black -> print_string "X"
                | Some White -> print_string "O"
            end;
            f := false;
            if x = tx && y = ty then (print_string "]"; f := true)
        done;
        f := false;
        print_newline ();
    done;
    flush Pervasives.stdout

(*** ̿ϥɥ ************************************************************)
(* version 1 ̿ϥɥ *)
let ver1_handler id idx in_c out_c =
    let wait () = flush out_c; ignore (input_line in_c) in
    {
        id = id;
        idx = idx;
        start_game = (fun oid -> ());
        end_game = (fun wid -> ());
        start_set = (fun clr rst -> Printf.fprintf out_c "START\n\n"; wait ());
        end_set = (fun slf opp -> Printf.fprintf out_c "END\n\n"; wait ());
        next =
            (fun brd clr hty rst ->
                Printf.fprintf
                    out_c
                    "NEXT\n%s %s\n"
                    (string_of_board brd)
                    (string_of_color clr);
                flush out_c;
                int_x_int_of_move (input_line in_c));
        close = (fun () -> close_in in_c; close_out out_c);
    }

(* version 2 ̿ϥɥ *)
let ver2_handler id idx in_c out_c =
    let wait () = flush out_c; ignore (input_line in_c) in
    {
        id = id;
        idx = idx;
        start_game = (fun oid ->
            Printf.fprintf out_c "START_GAME\n%s\n" oid; wait ());
        end_game = (fun wid ->
            Printf.fprintf out_c "END_GAME\n%s\n" wid; wait ());
        start_set = (fun clr rst ->
            Printf.fprintf out_c "START\n%s %d\n" (string_of_color clr) rst;
            wait ());
        end_set = (fun slf opp ->
            Printf.fprintf out_c "END\n%d %d\n" slf opp; wait ());
        next =
            (fun brd clr hty rst ->
                Printf.fprintf
                    out_c
                    "NEXT\n%s %s %s %d\n"
                    (string_of_board brd)
                    (string_of_color clr)
                    (string_of_history hty)
                    rst;
                flush out_c;
                int_x_int_of_move (input_line in_c));
        close = (fun () -> close_in in_c; close_out out_c);
    }

(*** ᥤץ ********************************************************)
(* 1 򤹤 *)
let play_set n hdl1 hdl2 time_limit =
    (* 1 ʤ *)
    let rec next brd (hdl, clr, rst) (ohdl, _, _ as ohdls) pass_flag hty =
        if can_put_anywhere brd clr then
        begin
            (* ޤ *)
            set_signal sigalrm (Signal_handle (fun _ -> hdl.close ()));
            let r = (float_of_int rst) /. 1000. in
            ignore (setitimer ITIMER_REAL { it_interval = 0.; it_value = r });
            (* ץ쥤䡼֤Ƥ餦 *)
            let t = gettimeofday () in
            let brd' = stone_list_list_of_board brd in
            let (x, y) = try hdl.next brd' clr hty rst with _ -> (-1, -1) in
            let t = int_of_float (((gettimeofday ()) -. t) *. 1000.) in
            (* ֤Ƥä *)
            if rst < t || x < 0 then
                (brd, x, y, Some ohdl, 64, 0, "time out") (* ڤ *)
            else
                try
                    (* ֤ *)
                    let brd' = put_stone brd clr x y in
                    if print_each_board then print_board brd' x y;
                    next
                        brd'
                        ohdls
                        (hdl, clr, rst - t)
                        false
                        ((clr, x, y)::hty)
                with CannotPutHere -> (* ʼ *)
                    (brd, x, y, Some ohdl, 64, 0, "illigal move")
        end
        else
            (* ֤ȤϤʤ *)
            if not pass_flag then
                next brd ohdls (hdl, clr, rst) true hty (* ѥ *)
            else
                (* ξԥѥʤΤǽλ *)
                let slf_r = ref 0 and opp_r = ref 0 in
                for y = 0 to 7 do
                    for x = 0 to 7 do
                        match get_stone brd x y with
                        | Some clr' when clr = clr' -> incr slf_r
                        | Some _ -> incr opp_r
                        | None -> ()
                    done
                done;
                let slf = !slf_r and opp = !opp_r in
                if slf = opp then (brd, -1, -1, None, slf, opp, "")
                else if slf > opp then (brd, -1, -1, Some hdl, slf, opp, "")
                else (brd, -1, -1, Some ohdl, opp, slf, "")
    in
    Printf.printf "Set %d start.\n" n;
    (*print_board init_board ~-1 ~-1;*)
    hdl1.start_set Black time_limit;
    hdl2.start_set White time_limit;
    flush Pervasives.stdout;
    let hdl1s = (hdl1, Black, time_limit)
    and hdl2s = (hdl2, White, time_limit) in
    let (brd, tx, ty, whdl, wc, lc, msg) =
        next init_board hdl1s hdl2s false []
    in
    if msg <> "" then print_endline ("Msg: " ^ msg);
    print_board brd tx ty;
    let (wid, w, c1, c2) =
        match whdl with
        | None -> ("even", 0, wc, lc)
        | Some whdl when whdl == hdl1 -> (hdl1.id, 1, wc, lc)
        | _ -> (hdl2.id, -1, lc, wc)
    in
    Printf.printf "Set %d end.\n" n;
    Printf.printf "Black %s(%d) vs White %s(%d)\n" hdl1.id c1 hdl2.id c2;
    Printf.printf "Winner : %s\n" wid;
    flush Pervasives.stdout;
    (try hdl1.end_set c1 c2 with _ -> ());
    (try hdl2.end_set c2 c1 with _ -> ());
    (w, c1, c2, msg = "time out")

(* 򤹤 *)
let play_game hdl1 hdl2 =
    Printf.printf "\nGame start.\n";
    flush Pervasives.stdout;
    hdl1.start_game hdl2.id;
    hdl2.start_game hdl1.id;
    let rec loop n (w1, w2, c1, c2 as rsl) =
        if n < number_of_sets / 2 then
            let f b = if b then 1 else 0 in
            let time_limit =
                if n < number_of_sets / 4 then time_limit_1 else time_limit_2
            in
            let (w, c1', c2', tout) =
                play_set (n * 2 + 1) hdl1 hdl2 time_limit
            in
            if tout then
                if w > 1 then (1, 0, 0, 0) else (0, 1, 0, 0)
            else
                let (w1, w2, c1, c2) =
                    (w1 + f (w > 0), w2 + f (w < 0), c1 + c1', c2 + c2')
                in
                let (w, c2', c1', tout) =
                    play_set (n * 2 + 2) hdl2 hdl1 time_limit
                in
                if tout then
                    if w > 1 then (0, 1, 0, 0) else (1, 0, 0, 0)
                else
                    loop
                        (n + 1)
                        (w1 + f (w < 0), w2 + f (w > 0), c1 + c1', c2 + c2')
        else
            rsl
    in
    let (w1, w2, c1, c2) = loop 0 (0, 0, 0, 0) in
    let wid =
        if w1 > w2 then hdl1.id else if w2 > w1 then hdl2.id
        else if c1 > c2 then hdl1.id else if c2 > c1 then hdl2.id
        else "even"
    in
    Printf.printf "Game end.\n\nResults...\n";
    if c1 > 0 || c2 > 0 then
    begin
        Printf.printf "%s : win %d (stone %d)\n" hdl1.id w1 c1;
        Printf.printf "%s : win %d (stone %d)\n" hdl2.id w2 c2;
    end
    else
        Printf.printf "Time out\n";
    Printf.printf "Winner : %s\n" wid;
    flush Pervasives.stdout;
    (try hdl1.end_game wid with _ -> ());
    (try hdl2.end_game wid with _ -> ());
    ()

(*** ץ̿Ω ********************************************************)
(* 2 ³Ԥ *)
let accept port =
    let servsock = socket PF_INET SOCK_STREAM 0 in
    setsockopt servsock SO_REUSEADDR true;
    bind servsock (ADDR_INET (inet_addr_any, port));
    listen servsock 2;
    Printf.printf "Listening : %d\n" port;
    flush Pervasives.stdout;
    let accept idx =
        let (fd, sa) = accept servsock in
        let in_c = in_channel_of_descr fd in
        let out_c = out_channel_of_descr fd in
        (in_c, out_c)
    in
    let c1 = accept 1 in
    let c2 = accept 2 in
    (try shutdown servsock SHUTDOWN_ALL with _ -> ());
    close servsock;
    (c1, c2)

(* 2 ץ *)
let exec p1 p2 =
    let exec p = open_process (p ^ " -c") in (exec p1, exec p2)

(* ư *)
let _ =
    let (c1, c2) =
        if process1 <> "" && process2 <> "" then
            exec process1 process2
        else
            accept port
    in
    let make_handle idx (in_c, out_c) =
        let line = String.uppercase (input_line in_c) in
        let (ver, id) = Scanf.sscanf line "%d %s" (fun ver id -> (ver, id)) in
        Printf.printf "Accepted %d : %s\n" idx id;
        flush Pervasives.stdout;
        match ver with
        | 1 -> ver1_handler id idx in_c out_c
        | 2 -> ver2_handler id idx in_c out_c
        | _ -> failwith "unsupported version."
    in
    let hdl1 = make_handle 1 c1 in
    let hdl2 = make_handle 2 c2 in
    if hdl1.id = hdl2.id then
    begin
        hdl1.id <- hdl1.id ^ "1";
        hdl2.id <- hdl2.id ^ "2";
    end;
    play_game hdl1 hdl2

(*

***** rvserver ưλ *****

- ̾ΥХ⡼
./rvserver -p <ݡ = 10000> -c <åȿ = 4>
ex) $ ./rvserver -p 12321 -c 100

- 󥽡⡼
./rvserver -p1 <ץ쥤䡼 1 ¹ԥեѥ> -p2 <ץ쥤䡼 2 ¹ԥեѥ>
ex) $ ./rvserver -p1 ./mak ./rvserver -p2 ./mak2

*)
